package org.prlab.idic.privacypolicy;

import android.bluetooth.*;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Bundle;
import android.support.v7.app.AppCompatActivity;
import android.support.v7.widget.RecyclerView;
import android.util.Log;
import android.view.View;
import android.widget.*;
import org.prlab.idic.privacypolicy.adapter.MessageRecyclerViewAdapter;
import org.prlab.idic.privacypolicy.ble.BLECustomScript;
import org.prlab.idic.privacypolicy.pojo.Message;
import org.prlab.idic.privacypolicy.schedule.Schedulable;
import org.prlab.idic.privacypolicy.storage.PrivacyPolicyStorageImpl;
import org.prlab.idic.privacypolicy.tool.BaseUtils;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;

import static android.bluetooth.BluetoothGatt.GATT_SUCCESS;

public class DeviceOperatorActivity extends AppCompatActivity {

    private static final String TAG = DeviceOperatorActivity.class.getSimpleName();
    private UUID PRIVACY_POLICY_ACCEPT_UUID;

    private PrivacyPolicyStorageImpl mPolicyStorage;
    private String mTargetDeviceAddress;
    private List<Message> mMessages = new LinkedList<>();

    private TextView mMessageBarTextView;
    private EditText mTemperatureEditText;
    private ImageButton mTemperatureButton;
    private Switch mAirPurifierSwitch;
    private RecyclerView.Adapter mMessageRecyclerViewAdapter;

    private BluetoothManager mBluetoothManager;
    private BluetoothAdapter mBluetoothAdapter;
    private BluetoothGatt mBluetoothGatt;
    private BroadcastReceiver mPairRequestReceiver;
    private BroadcastReceiver mBondStateReceiver;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_device_operator);

        mTargetDeviceAddress = getIntent().getStringExtra(getString(R.string.intent_target_address_key));
        mPolicyStorage = new PrivacyPolicyStorageImpl(getContentResolver(), mTargetDeviceAddress);
        PRIVACY_POLICY_ACCEPT_UUID = UUID.fromString(getString(R.string.uuid_accept_privacy_policy));

        initView();
        initBluetooth();
    }

    private void initView() {
        mMessageBarTextView = (TextView) findViewById(R.id.text_operation_message_bar);
        mTemperatureEditText = (EditText) findViewById(R.id.edit_operation_current_temperature);
        mTemperatureButton = (ImageButton) findViewById(R.id.button_operation_send_temperature);
        mTemperatureButton.setOnClickListener(mTemperatureClickedListener);

        mAirPurifierSwitch = (Switch) findViewById(R.id.switch_operation_air_purifier);
        mAirPurifierSwitch.setOnCheckedChangeListener(mOnAirPurifierSwitchCheckedListener);

        RecyclerView mMessageRecyclerView = (RecyclerView) findViewById(R.id.recycler_view_operation_info);
        mMessageRecyclerViewAdapter = new MessageRecyclerViewAdapter(getResources(), mMessages);
        mMessageRecyclerView.setAdapter(mMessageRecyclerViewAdapter);
    }

    private void initBluetooth() {
        mBluetoothManager = (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        if (mBluetoothManager != null) {
            mBluetoothAdapter = mBluetoothManager.getAdapter();
            if (mBluetoothAdapter != null)
                connect();
        }
    }

    private final UUID ROOT_UUID = UUID.fromString("01122334-4556-6778-899a-abbccddeeff0");
    private final UUID TEMPERATURE_UUID = UUID.fromString("01122334-4556-6778-899a-abbccddeeff1");
    private final UUID AIR_UUID = UUID.fromString("01122334-4556-6778-899a-abbccddeeff2");

    private View.OnClickListener mTemperatureClickedListener = new View.OnClickListener() {
        @Override
        public void onClick(View v) {
            if (mBluetoothGatt != null) {
                BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
                BluetoothGattCharacteristic temperature = service == null ? null : service.getCharacteristic(TEMPERATURE_UUID);
                if (temperature != null) {
                    temperature.setValue(BaseUtils.int2Bytes(Integer.parseInt(String.valueOf(mTemperatureEditText.getText()))));
                    mBluetoothGatt.writeCharacteristic(temperature);
                }
            }
        }
    };

    private CompoundButton.OnCheckedChangeListener mOnAirPurifierSwitchCheckedListener = new CompoundButton.OnCheckedChangeListener() {
        @Override
        public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
            if (mBluetoothGatt != null) {
                BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
                BluetoothGattCharacteristic airPurifier = service == null ? null : service.getCharacteristic(AIR_UUID);
                if (airPurifier != null) {
                    airPurifier.setValue(isChecked ? BaseUtils.int2Bytes(1) : BaseUtils.int2Bytes(0));
                    mBluetoothGatt.writeCharacteristic(airPurifier);
                }
            }
        }
    };

    private List<Schedulable> mPermissionSettingQueue = new LinkedList<>();

    private BLECustomScript mBLECustomScript = new BLECustomScript() {

        int expectedSuccessValue = 0;
        int successValue = 0;

        private void requestServices() {
            if (successValue == expectedSuccessValue) {
                setMessageBar("Try to request services.");
                addMessage("Request Services", "Try to get temperature & air purifier state.");
                requestTemperature();
            }
        }

        private void requestValue(UUID characteristicUUID) {
            synchronized (mBluetoothGatt) {
                BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
                BluetoothGattCharacteristic characteristic = service.getCharacteristic(characteristicUUID);
                if (characteristic != null) {
                    mBluetoothGatt.readCharacteristic(characteristic);
                }
            }
        }

        private void requestTemperature() {
            requestValue(TEMPERATURE_UUID);
        }

        private void requestAirPurifier() {
            requestValue(AIR_UUID);
        }

        private void resolveTemperature(final int value) {
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    if (value == -99) {
                        addMessage("Temperature Service", "Get failure -> " + value);
                        mTemperatureEditText.setEnabled(false);
                        mTemperatureButton.setEnabled(false);
                    } else {
                        addMessage("Temperature Service", "Get success -> " + value);
                        mTemperatureEditText.setEnabled(true);
                        mTemperatureButton.setEnabled(true);
                    }
                    mTemperatureEditText.setText(String.valueOf(value));
                }
            });
        }

        private void resolveAirPurifier(final int value) {
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    if (value != -1) {
                        addMessage("Air Purifier Service", "Get success -> " + (value == 1 ? "running" : "sleeping"));
                    } else {
                        addMessage("Air Purifier Service", "Get failure -> " + String.valueOf(value));
                    }
                    mAirPurifierSwitch.setChecked(value == 1);
                }
            });
        }

        private boolean settingPermission() {
            if (mPermissionSettingQueue.size() > 0) {
                Log.i(TAG, "Continue to work.");
                mPermissionSettingQueue.remove(0).work();
                return false;
            }
            return true;
        }

        @Override
        public void onDiscoverServiceCompleted(final BluetoothGatt gatt) {
            expectedSuccessValue = 0;
            successValue = 0;

            setMessageBar("Enable privacy policy setting.");
            for (BluetoothGattService service : gatt.getServices()) {
                List<String> layerUUIDs = new LinkedList<>();
                layerUUIDs.add(String.valueOf(service.getUuid()));
                //Write service layer selection
                int selection = mPolicyStorage.get(layerUUIDs);
                if (selection == -1) selection = 0;
                final BluetoothGattCharacteristic characteristic = service.getCharacteristic(PRIVACY_POLICY_ACCEPT_UUID);
                if (characteristic != null) {
                    final int finalSelection = selection;
                    mPermissionSettingQueue.add(new Schedulable() {
                        @Override
                        public void work() {
                            characteristic.setValue(BaseUtils.int2Bytes(finalSelection));
                            Log.i(TAG, Arrays.toString(BaseUtils.int2Bytes(finalSelection)));
                            gatt.writeCharacteristic(characteristic);
                            expectedSuccessValue++;
                        }
                    });
                    addMessage("Policy Setting", "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) + " to " + selection + ".");
                }
                for (BluetoothGattCharacteristic characteristic2 : service.getCharacteristics()) {
                    layerUUIDs.add(String.valueOf(characteristic2.getUuid()));
                    //Write characteristic layer selection
                    selection = mPolicyStorage.get(layerUUIDs);
                    if (selection == -1) selection = 0;
                    final BluetoothGattDescriptor descriptor = characteristic2.getDescriptor(PRIVACY_POLICY_ACCEPT_UUID);
                    if (descriptor != null) {
                        final int finalSelection1 = selection;
                        mPermissionSettingQueue.add(new Schedulable() {
                            @Override
                            public void work() {
                                descriptor.setValue(BaseUtils.int2Bytes(finalSelection1));
                                Log.i(TAG, Arrays.toString(BaseUtils.int2Bytes(finalSelection1)));
                                gatt.writeDescriptor(descriptor);
                                expectedSuccessValue++;
                            }
                        });
                        addMessage("Policy Setting", "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) + " to " + selection + ".");
                    }
                    layerUUIDs.remove(layerUUIDs.size() - 1);
                }
            }
            settingPermission();
        }

        @Override
        public void onWriteCharacteristicCompleted(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status, byte[] values) {
            if (PRIVACY_POLICY_ACCEPT_UUID.equals(characteristic.getUuid())) {
                successValue++;
                List<String> layerUUIDs = new LinkedList<>();
                layerUUIDs.add(characteristic.getService().getUuid().toString());
                addMessage("Policy Setting",
                        "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) + (status == GATT_SUCCESS ? " success" : " failure") + " .");
                if (settingPermission())
                    requestServices();
            } else if (TEMPERATURE_UUID.equals(characteristic.getUuid())) {
                addMessage("Temperature Setting", "Set " + BaseUtils.bytesToInt(values) + " to the air conditioning.");
                requestTemperature();
            } else if (AIR_UUID.equals(characteristic.getUuid())) {
                addMessage("Air Purifier Setting", "Turn " + (BaseUtils.bytesToInt(values) == 1 ? "on " : "off ") + "the air purifier" + ".");
                requestAirPurifier();
            }
        }

        @Override
        public void onReadCharacteristicCompleted(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status, byte[] values) {
            if (TEMPERATURE_UUID.equals(characteristic.getUuid())) {
                resolveTemperature(BaseUtils.bytesToInt(values));
                if (mTemperatureEditText.getText().toString().equals("")) {
                    requestAirPurifier();
                }
            } else if (AIR_UUID.equals(characteristic.getUuid())) {
                resolveAirPurifier(BaseUtils.bytesToInt(values));
            }
        }

        @Override
        public void onWriteDescriptorCompleted(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status, byte[] values) {
            if (PRIVACY_POLICY_ACCEPT_UUID.equals(descriptor.getUuid())) {
                successValue++;
                List<String> layerUUIDs = new LinkedList<>();
                layerUUIDs.add(descriptor.getCharacteristic().getService().getUuid().toString());
                layerUUIDs.add(descriptor.getCharacteristic().getUuid().toString());
                addMessage("Policy Setting", "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) +
                        (status == GATT_SUCCESS ? " success" : " failure") + " .");
                if (settingPermission())
                    requestServices();
            }
        }

        @Override
        public void onReadDescriptorCompleted(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status, byte[] values) {

        }

        @Override
        public void onPairRequestReceived(BluetoothDevice device, int pin) {

        }

        @Override
        public void onPaired(BluetoothDevice device) {

        }

        @Override
        public void onUnpaired(BluetoothDevice device) {

        }

    };

    private BluetoothGattCallback mBluetoothGattCallback = new BluetoothGattCallback() {

        @Override
        public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
            String mac = gatt.getDevice().getAddress();
            if (status == GATT_SUCCESS) {
                if (newState == BluetoothGatt.STATE_CONNECTED) {
                    gatt.discoverServices();
                    addMessage("Connection State Change ", "Device " + mac + " is connected.");
                } else if (newState == BluetoothGatt.STATE_DISCONNECTED) {
                    mBluetoothGatt.close();
                    mBluetoothGatt = null;
                    addMessage("Connection State Change ", "Device " + mac + " is disconnected.");
                }
            }
            //重連機制，但好像只會一直bind住。
            else if (status == 133 && newState == BluetoothGatt.STATE_DISCONNECTED) {
                mBluetoothGatt.close();
                connect();
            }
        }

        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            if (status == GATT_SUCCESS) {
                mBLECustomScript.onDiscoverServiceCompleted(gatt);
            }
        }

        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
//            if ((status & (BluetoothGatt.GATT_INSUFFICIENT_AUTHENTICATION | BluetoothGatt.GATT_INSUFFICIENT_ENCRYPTION)) != 0) {
//                boolean isSuccessful = gatt.getDevice().createBond();
//                Log.i(TAG, "Create Bond is " + isSuccessful);
//            }
            mBLECustomScript.onReadCharacteristicCompleted(mBluetoothGatt, characteristic, status, characteristic.getValue());

        }

        @Override
        public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            mBLECustomScript.onWriteCharacteristicCompleted(mBluetoothGatt, characteristic, status, characteristic.getValue());
        }


        @Override
        public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            mBLECustomScript.onReadDescriptorCompleted(mBluetoothGatt, descriptor, status, descriptor.getValue());
        }

        @Override
        public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            mBLECustomScript.onWriteDescriptorCompleted(mBluetoothGatt, descriptor, status, descriptor.getValue());
        }
    };

    private void registerPairRequestReceiver() {
        IntentFilter intentFilter = new IntentFilter(BluetoothDevice.ACTION_PAIRING_REQUEST);
        BroadcastReceiver receiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                if (intent.getAction().equals(BluetoothDevice.ACTION_PAIRING_REQUEST)) {
                    Log.d(TAG, "PIN " + intent.getIntExtra(BluetoothDevice.EXTRA_PAIRING_KEY, 0));
                    BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                    int pin = intent.getIntExtra(BluetoothDevice.EXTRA_PAIRING_KEY, 0);
                    mBLECustomScript.onPairRequestReceived(device, pin);
                }
            }
        };
        registerReceiver(receiver, intentFilter);
        mPairRequestReceiver = receiver;
    }

    private void unregisterPairRequestReceiver() {
        if (mPairRequestReceiver != null) {
            unregisterReceiver(mPairRequestReceiver);
            mPairRequestReceiver = null;
        }
    }

    private void registerBondStateReceiver() {
//        mMessageCallback.displayMessage("Pair", "Ready to pair " + mTargetAddress, true);
        IntentFilter intentFilter = new IntentFilter(BluetoothDevice.ACTION_BOND_STATE_CHANGED);
        BroadcastReceiver receiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                String action = intent.getAction();
                if (BluetoothDevice.ACTION_BOND_STATE_CHANGED.equals(action)) {
                    final int state = intent.getIntExtra(BluetoothDevice.EXTRA_BOND_STATE, BluetoothDevice.ERROR);
                    final int preState = intent.getIntExtra(BluetoothDevice.EXTRA_PREVIOUS_BOND_STATE, BluetoothDevice.ERROR);
                    Log.i(TAG, "Bond [" + preState + " to " + state + "].");
                    BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                    if (state == BluetoothDevice.BOND_BONDED && preState == BluetoothDevice.BOND_BONDING) {
                        mBLECustomScript.onPaired(device);
                    } else if (state == BluetoothDevice.BOND_NONE && preState == BluetoothDevice.BOND_BONDED) {
                        mBLECustomScript.onUnpaired(device);
                    }
                }

            }
        };
        registerReceiver(receiver, intentFilter);
        mBondStateReceiver = receiver;
    }

    private void unregisterBondStateReceiver() {
        if (mBondStateReceiver != null) {
            unregisterReceiver(mBondStateReceiver);
            mBondStateReceiver = null;
        }
    }

    private boolean connect() {
        if (mBluetoothAdapter == null) return false;
        BluetoothDevice device = mBluetoothAdapter.getRemoteDevice(mTargetDeviceAddress);

        int connectionState = mBluetoothManager.getConnectionState(device, BluetoothProfile.GATT);
        Log.i(TAG, "Device connection status is " + connectionState + ".");

        if (device != null && connectionState == BluetoothProfile.STATE_DISCONNECTED) {
            mBluetoothGatt = device.connectGatt(this, false, mBluetoothGattCallback);
            return mBluetoothGatt != null;
        } else {
            return false;
        }
    }


    private void disconnect() {
        if (mBluetoothGatt != null) {
            int connectionState = mBluetoothManager.getConnectionState(mBluetoothGatt.getDevice(), BluetoothProfile.GATT);
            Log.i(TAG, "Device connection status is " + connectionState + ".");
            if (connectionState == BluetoothProfile.STATE_CONNECTED) {
                boolean isBonded = mBluetoothAdapter.getBondedDevices().contains(mBluetoothGatt.getDevice());
                Log.i(TAG, "Device is already bonded. :" + isBonded);
//                try {
//                    Method method = mBluetoothGatt.getDevice().getClass().getMethod("removeBond", (Class[]) null);
//                    method.invoke(mBluetoothGatt.getDevice(), (Object[]) null);
//                } catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException e) {
//                    e.printStackTrace();
//                }
                mBluetoothGatt.disconnect();
            }
        }
    }

    private void setMessageBar(final String message) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                mMessageBarTextView.setText(message);
            }
        });
    }

    private void addMessage(String event, String message) {
        addMessage(event, message, false);
    }

    private void addMessage(final String event, final String message, final boolean isMarked) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                mMessages.add(new Message()
                        .setEvent(event)
                        .setContent(message)
                        .setMarked(isMarked));
                mMessageRecyclerViewAdapter.notifyItemInserted(mMessages.size() - 1);
            }
        });
    }

    @Override
    public void onBackPressed() {
        super.onBackPressed();
        if (mBluetoothGatt != null) {
            mBluetoothGatt.disconnect();
            mBluetoothGatt.close();
        }
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        if (mBluetoothGatt != null) {
            mBluetoothGatt.disconnect();
            mBluetoothGatt.close();
        }
    }
}
