package org.prlab.idic.privacypolicy;

import android.Manifest;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothManager;
import android.bluetooth.le.*;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.graphics.drawable.Drawable;
import android.os.Build;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.support.v4.app.ActivityCompat;
import android.support.v7.app.AppCompatActivity;
import android.support.v7.widget.RecyclerView;
import android.util.Log;
import android.util.SparseArray;
import android.view.Menu;
import android.view.MenuItem;
import android.widget.TextView;
import android.widget.Toast;
import org.prlab.idic.privacypolicy.adapter.ScanResultRecyclerViewAdapter;
import org.prlab.idic.privacypolicy.ble.BluetoothDeviceInformation;
import org.prlab.idic.privacypolicy.tool.beacon.BeaconType;
import org.prlab.idic.privacypolicy.tool.beacon.BeaconUtils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

public class MainActivity extends AppCompatActivity {

    private static final int PERMISSION_CENTRAL_REQUEST_CODE = 87;
    private static final String TAG = MainActivity.class.getSimpleName();

    private TextView mMessageBarTextView;
    private RecyclerView mScanResultRecyclerView;
    private RecyclerView.Adapter mScanResultRecyclerViewAdapter;

    private List<BluetoothDeviceInformation> mInformationList = new ArrayList<>();

    private boolean isScanning = false;

    private BluetoothManager mBluetoothManager;
    private BluetoothAdapter mBluetoothAdapter;
    private BluetoothLeScanner mBluetoothLeScanner;
    private ScanSettings.Builder mScanSettingsBuilder = new ScanSettings.Builder()
            .setScanMode(ScanSettings.SCAN_MODE_BALANCED);

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        initView();
        setupPermissions();
    }

    private void initBluetooth() {
        mBluetoothManager = (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        if (mBluetoothManager != null) {
            mBluetoothAdapter = mBluetoothManager.getAdapter();
            if (mBluetoothAdapter != null) {
                mBluetoothLeScanner = mBluetoothAdapter.getBluetoothLeScanner();
            }
        }
    }

    private void initView() {
        mMessageBarTextView = (TextView) findViewById(R.id.text_main_message_bar);

        displayMessage(getString(R.string.message_press_scan_first));

        mScanResultRecyclerView = (RecyclerView) findViewById(R.id.recycler_view_main_result);
        mScanResultRecyclerViewAdapter = new ScanResultRecyclerViewAdapter(getResources(), mOnDisplayPolicyClickListener, mOnConnectDeviceClickListener, mInformationList);
        mScanResultRecyclerView.setAdapter(mScanResultRecyclerViewAdapter);
    }

    private ScanResultRecyclerViewAdapter.ListItemClickListener mOnDisplayPolicyClickListener = new ScanResultRecyclerViewAdapter.ListItemClickListener() {
        @Override
        public void onListItemClick(int clickedItemIndex) {
            BluetoothDeviceInformation device = mInformationList.get(clickedItemIndex);
            if (null != device.getReportUUID()) {
                Intent intent = new Intent();
                intent.setClass(MainActivity.this, ReportActivity.class);
                intent.putExtra(getString(R.string.intent_report_id_key), device.getReportUUID());
                intent.putExtra(getString(R.string.intent_report_mac_key), device.getAddress());
                startActivity(intent);
            }
        }
    };

    private ScanResultRecyclerViewAdapter.ListItemClickListener mOnConnectDeviceClickListener = new ScanResultRecyclerViewAdapter.ListItemClickListener() {
        @Override
        public void onListItemClick(int clickedItemIndex) {
            BluetoothDeviceInformation device = mInformationList.get(clickedItemIndex);
            if (null != device.getReportUUID()) {
                Intent intent = new Intent();
                intent.setClass(MainActivity.this, DeviceOperatorActivity.class);
                intent.putExtra(getString(R.string.intent_target_address_key), device.getAddress());
                startActivity(intent);
            }
        }
    };

    private static String[] permissionsWeNeed = new String[]{
            Manifest.permission.INTERNET,
            Manifest.permission.BLUETOOTH_ADMIN,
            Manifest.permission.BLUETOOTH,
            Manifest.permission.ACCESS_FINE_LOCATION,
            Manifest.permission.ACCESS_COARSE_LOCATION};

    private void setupPermissions() {
        // If we don't have the record audio permission...
        boolean isGranted = true;
        for (String permission : permissionsWeNeed) {
            isGranted &= ActivityCompat.checkSelfPermission(this, permission) == PackageManager.PERMISSION_GRANTED;
        }

        if (!isGranted) {
            // And if we're on SDK M or later...
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                // Ask again, nicely, for the permissions.
                requestPermissions(permissionsWeNeed, PERMISSION_CENTRAL_REQUEST_CODE);
            }
        } else {
            // Otherwise, permissions were granted and we are ready to go!
            initBluetooth();
        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String permissions[], @NonNull int[] grantResults) {
        switch (requestCode) {
            case PERMISSION_CENTRAL_REQUEST_CODE: {
                // If request is cancelled, the result arrays are empty.
                boolean isGranted = grantResults.length > 0;
                for (int grantResult : grantResults) {
                    isGranted &= grantResult == PackageManager.PERMISSION_GRANTED;
                }
                if (isGranted) {
                    initBluetooth();
                } else {
                    Toast.makeText(this, "Permission for BLE central device not granted. Central device APP can't run.", Toast.LENGTH_LONG).show();
                    finish();
                }
            }
        }
    }


    /**
     * Methods for setting up the menu
     **/
    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        getMenuInflater().inflate(R.menu.menu_main, menu);
        menu.getItem(0).setIcon(isScanning ? R.mipmap.ic_bluetooth_disabled : R.mipmap.ic_bluetooth_searching);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        int id = item.getItemId();
        if (id == R.id.action_search) {
            Drawable drawable = getDrawable(R.mipmap.ic_bluetooth_searching);
            if (item.getIcon() != null && item.getIcon().getConstantState() != null && drawable != null) {
                if (item.getIcon().getConstantState().equals(drawable.getConstantState())) {
                    displayMessage(getString(R.string.message_scan_start));
                    if (startScan()) {
                        item.setIcon(R.mipmap.ic_bluetooth_disabled);
                    } else {
                        if (!mBluetoothAdapter.isEnabled()) {
                            displayMessage(getString(R.string.message_enable_first));
                            if (mBluetoothAdapter.enable())
                                initBluetooth();
                        }
                    }
                } else {
                    if (stopScan()) {
                        displayMessage(getString(R.string.message_scan_stop));
                        item.setIcon(R.mipmap.ic_bluetooth_searching);
                    }
                }
            }
        }
        return super.onOptionsItemSelected(item);
    }

    private ScanCallback mLeScanCallback = new ScanCallback() {

        @Override
        public void onScanResult(int callbackType, ScanResult result) {
            BluetoothDeviceInformation information = toBluetoothDeviceInformation(result);
            int index = mInformationList.indexOf(information);
            if (index != -1) {
                mInformationList.set(index, information);
                mScanResultRecyclerViewAdapter.notifyItemChanged(index);
            } else {
                mInformationList.add(information);
                mScanResultRecyclerViewAdapter.notifyItemInserted(mInformationList.size() - 1);
            }
        }

        @Override
        public void onBatchScanResults(List<ScanResult> results) {
            for (ScanResult result : results) {
                onScanResult(0, result);
            }
        }

        @Override
        public void onScanFailed(int errorCode) {
            Log.e(TAG, "Failed to scan : " + errorCode);
        }

    };

    private static BluetoothDeviceInformation toBluetoothDeviceInformation(ScanResult result) {
        SparseArray<byte[]> manufacturerSpecificDataSparseArray = result.getScanRecord().getManufacturerSpecificData();
        byte[] manufacturerData = null;
        for (int i = 0; i < manufacturerSpecificDataSparseArray.size(); i++) {
            int key = manufacturerSpecificDataSparseArray.keyAt(0);
            manufacturerData = manufacturerSpecificDataSparseArray.get(key);
        }
        String uuid = null;
        BluetoothDevice device = result.getDevice();
        if (null != manufacturerData && BeaconType.NORMAL_BEACON == BeaconUtils.getBeaconType(manufacturerData)) {
            uuid = BeaconUtils.calculateUuidString(Arrays.copyOfRange(manufacturerData, 2, 18));
        }
        return new BluetoothDeviceInformation()
                .setName(device == null ? null : device.getName())
                .setAddress(device == null ? null : device.getAddress())
                .setRssi(result.getRssi())
                .setBonded(device != null && device.getBondState() == BluetoothDevice.BOND_BONDED)
                .setReportUUID(uuid);
    }

    private boolean startScan() {
        if (mBluetoothLeScanner != null) {
            mBluetoothLeScanner.startScan(new LinkedList<ScanFilter>(), mScanSettingsBuilder.build(), mLeScanCallback);
            isScanning = true;
            return true;
        }
        return false;
    }

    private boolean stopScan() {
        if (mBluetoothLeScanner != null) {
            mBluetoothLeScanner.stopScan(mLeScanCallback);
            isScanning = false;
            return true;
        }
        return false;
    }

    private void displayMessage(String message) {
        mMessageBarTextView.setText(message);
    }

}
