package org.prlab.idic.privacypolicy.storage;

import android.content.ContentResolver;
import android.content.ContentValues;
import android.database.Cursor;
import android.net.Uri;
import android.util.Log;
import org.prlab.idic.privacypolicy.tool.BaseUtils;

import java.util.LinkedList;
import java.util.List;

/**
 * Created by IDIC on 2017/2/15.
 */
public class PrivacyPolicyStorageImpl implements StorageInterface<Integer, String> {

    private static final String TAG = PrivacyPolicyStorageImpl.class.getSimpleName();

    private ContentResolver mContentResolver;
    private String mTargetDeviceMAC;

    private static final String[] PRIVACY_POLICY_PROJECTION = {
            PrivacyPolicyContract.PolicyEntry.COLUMN_MAC,
            PrivacyPolicyContract.PolicyEntry.COLUMN_UUIDs,
            PrivacyPolicyContract.PolicyEntry.COLUMN_PREFERENCE
    };

    public static final int INDEX_POLICY_MAC = 0;
    public static final int INDEX_POLICY_UUIDs = 1;
    public static final int INDEX_POLICY_PREFERENCE = 2;

    public PrivacyPolicyStorageImpl(ContentResolver contentResolver, String targetDeviceMAC) {
        this.mContentResolver = contentResolver;
        this.mTargetDeviceMAC = BaseUtils.shortMAC(targetDeviceMAC);
    }

    @Override
    public StorageInterface remove(List<String> layerUUIDs) {
        remove(BaseUtils.stringList2ArrayString(layerUUIDs));
        return this;
    }

    private void remove(String layerUUIDs) {
        int impactRow = mContentResolver.delete(
                PrivacyPolicyContract.PolicyEntry.CONTENT_URI,
                PrivacyPolicyContract.getSqlSelectForMACAndUUIDsOnwards(),
                new String[]{mTargetDeviceMAC, layerUUIDs}
        );
        Log.i(TAG, "Remove [" + layerUUIDs + "] " + (impactRow == 1 ? "success" : "failure") + ".");
    }

    @Override
    public StorageInterface put(Integer selection, List<String> layerUUIDs) {
        put(selection, BaseUtils.stringList2ArrayString(layerUUIDs));
        return this;
    }

    private void put(Integer selection, String layerUUIDs) {
        ContentValues updateContentValues = new ContentValues();
        updateContentValues.put(PrivacyPolicyContract.PolicyEntry.COLUMN_PREFERENCE, selection);

        ContentValues insertContentValues = new ContentValues(updateContentValues);
        insertContentValues.put(PrivacyPolicyContract.PolicyEntry.COLUMN_MAC, mTargetDeviceMAC);
        insertContentValues.put(PrivacyPolicyContract.PolicyEntry.COLUMN_UUIDs, layerUUIDs);
        Uri mNewUri = mContentResolver.insert(
                PrivacyPolicyContract.PolicyEntry.CONTENT_URI,
                insertContentValues
        );
        if (mNewUri == null) {
            int count = mContentResolver.update(
                    PrivacyPolicyContract.PolicyEntry.CONTENT_URI,
                    updateContentValues,
                    PrivacyPolicyContract.getSqlSelectForMACAndUUIDsOnwards(),
                    new String[]{mTargetDeviceMAC, layerUUIDs}
            );
            Log.i(TAG, "Update " + String.valueOf(selection) + " to [" + layerUUIDs + "] " + (count == 1 ? "success" : "failure") + ".");
        } else {
            Log.i(TAG, "Put " + String.valueOf(selection) + " to [" + layerUUIDs + "] success.");
        }
        List<String> subUUIDs = getRelatedUUIDs(false, layerUUIDs);
        for (String subUUID : subUUIDs) {
            remove(subUUID);
        }
    }

    @Override
    public Integer get(List<String> layerUUIDs) {
        return get(BaseUtils.stringList2ArrayString(layerUUIDs));
    }

    private Integer get(String UUIDs) {
        try (Cursor cursor = mContentResolver.query(
                PrivacyPolicyContract.PolicyEntry.CONTENT_URI,
                PRIVACY_POLICY_PROJECTION,
                PrivacyPolicyContract.getSqlSelectForMACAndUUIDsOnwards(),
                new String[]{mTargetDeviceMAC, UUIDs},
                PrivacyPolicyContract.PolicyEntry.COLUMN_UUIDs + " ASC")) {

            if (cursor.getCount() == 1) {
                cursor.moveToFirst();
                String target = cursor.getString(INDEX_POLICY_PREFERENCE);

                Log.i(TAG, "Get [" + UUIDs + "] -> [" + String.valueOf(target) + "]");

                if (target == null) return null;
                return Integer.valueOf(target);
            }
            Log.i(TAG, "Get [" + UUIDs + "] -> null");
            return -1;
        }
    }

    private List<String> getRelatedUUIDs(boolean isLower, String UUIDs) {
        List<String> UUIDList = new LinkedList<>();
        if (isLower) {
            try (Cursor cursor = mContentResolver.query(
                    PrivacyPolicyContract.PolicyEntry.CONTENT_URI,
                    PRIVACY_POLICY_PROJECTION,
                    PrivacyPolicyContract.getSqlSelectForMACAndRelatedUUIDsOnwards(),
                    new String[]{mTargetDeviceMAC, UUIDs + "_%"},
                    PrivacyPolicyContract.PolicyEntry.COLUMN_UUIDs + " ASC")) {
                while (cursor.moveToNext()) {
                    String targetUUIDs = cursor.getString(INDEX_POLICY_UUIDs);
                    UUIDList.add(targetUUIDs);
                }
            }
        } else {
            if (UUIDs.contains(", ")) {
                UUIDs = UUIDs.substring(0, UUIDs.lastIndexOf(", "));
                try (Cursor cursor = mContentResolver.query(
                        PrivacyPolicyContract.PolicyEntry.CONTENT_URI,
                        PRIVACY_POLICY_PROJECTION,
                        PrivacyPolicyContract.getSqlSelectForMACAndRelatedUUIDsOnwards(),
                        new String[]{mTargetDeviceMAC, "%" + UUIDs},
                        PrivacyPolicyContract.PolicyEntry.COLUMN_UUIDs + " ASC")) {
                    while (cursor.moveToNext()) {
                        String targetUUIDs = cursor.getString(INDEX_POLICY_UUIDs);
                        UUIDList.add(targetUUIDs);
                    }
                }
            }
        }
        return UUIDList;
    }

}
